@@ -27,10 +27,16 @@ import codecs
 
 import string
 
+import re
+
 from storage_exceptions import *
 
 from pairtree_object import PairtreeStorageObject
 
+import pairtree_path as ppath
+
+import hashlib
+
 class PairtreeStorageClient(object):
     """A client that oversees the implementation of the Pairtree FS specification
     version 0.1.
@@ -59,7 +65,7 @@ class PairtreeStorageClient(object):
     Also, if you try to create a store over a directory that already exists, but which isn't
     a pairtree store that it can recognise, it will raise a L{NotAPairtreeStoreException}.
     """
-    def __init__(self, uri_base, store_dir, shorty_length):
+    def __init__(self, uri_base, store_dir, shorty_length, hashing_type=None):
         """
         Constructor
         @param store_dir: The file directory where the pairtree store is
@@ -68,6 +74,8 @@ class PairtreeStorageClient(object):
         @type uri_base: A URI fragment, like "http://example.org/"
         @param shorty_length: The size of the shorties in the pairtree implementation (Default: 2)
         @type shorty_length: integer
+        @param hashing_type: The name of the algorithm to use when hashing files, if left as None, this is disabled.
+        @type hashing_type: Any supported by C{hashlib}
         """
         self.store_dir = store_dir
         self.pairtree_root = os.path.join(self.store_dir, 'pairtree_root')
@@ -75,8 +83,19 @@ class PairtreeStorageClient(object):
         if uri_base:
             self.uri_base = uri_base
         self.shorty_length = shorty_length
+        self.hashing_type = hashing_type
+        # regexes
+        self._encode = re.compile(r"[\"*+,<=>?\\^|]|[^\x21-\x7e]", re.U)
+        self._decode = re.compile(r"\^(..)", re.U)
+        
         self._init_store()
 
+    def __char2hex(self, m):
+        return ppath.char2hex(m)
+
+    def __hex2char(self, m):
+        return ppath.hex2char(m)
+
     def id_encode(self, id):
         """
         The identifier string is cleaned of characters that are expected to occur rarely
@@ -123,34 +142,7 @@ class PairtreeStorageClient(object):
         @type id: identifier
         @returns: A string of the encoded identifier
         """
-        multichar_mapping = {u'"':u'^22',
-                             u'<':u'^3c',
-                             u'?':u'^3f',
-                             u'*':u'^2a',
-                             u'=':u'^3d',
-                             u'^':u'^5e',
-                             u'+':u'^2b',
-                             u'>':u'^3e',
-                             u'|':u'^7c',
-                             u',':u'^2c'
-                            }
-        second_pass_m = {u'/':u'=',
-                         u':':u'+',
-                         u'.':u','
-                        }
-        new_id = []
-        for char in id:
-            new_id.append(multichar_mapping.get(char, char))
-        # 2nd pass
-        # Ditched using .translate and string.maketrans due to weird and
-        # odd errors
-        second_pass = []
-        for char in u"".join(new_id):
-            second_pass.append(second_pass_m.get(char, char))
-        return u"".join(second_pass)
-        #        m = string.maketrans(u'/:.',u'=+,')
-        #        return foo.translate(m)
-        # return "".join(new_id).translate(string.maketrans('/:.','=+,'))
+        return ppath.id_encode(id)
 
     def id_decode(self, id):
         """
@@ -160,32 +152,7 @@ class PairtreeStorageClient(object):
         @type id: identifier
         @returns: A string of the decoded identifier
         """
-        multichar_mapping = {u'22':u'"',
-                             u'3c':u'<',
-                             u'3f':u'?',
-                             u'2a':u'*',
-                             u'3d':u'=',
-                             u'5e':u'^',
-                             u'2b':u'+',
-                             u'3e':u'>',
-                             u'7c':u'|',
-                             u'2c':u','
-                            }
-        dec_id = id.translate(string.maketrans(u'=+,',u'/:.'))
-        index = 0
-        new_id = []
-        while index < len(dec_id):
-            if dec_id[index] == '^':
-                code = "".join(dec_id[index+1:index+3]).lower()
-                if code in multichar_mapping:
-                    new_id.append(multichar_mapping[code])
-                    index = index + 2
-                else:
-                    raise Exception("Unknown character code found.")
-            else:
-                new_id.append(dec_id[index])
-            index = index +1
-        return "".join(new_id)
+        return ppath.id_decode(id)
 
     def _get_id_from_dirpath(self, dirpath):
         """
@@ -313,11 +280,10 @@ class PairtreeStorageClient(object):
 
         Currently, it ignores the possibility of a split end being
         'shielded' by a /obj/ folder
+        
+        Returns a generator, not a plain list since version 0.4.12
 
-        (Note - this will shortly be turned into a generator, due to the expense of
-        the call)
-
-        @returns: L{list}
+        @returns: L{generator}
         """
 
         objects = set()
@@ -326,16 +292,17 @@ class PairtreeStorageClient(object):
         if paths:
             d = paths.pop()
         while d:
-            for t in [x for x in os.listdir(d) if os.path.isdir(os.path.join(d, x))]:
+            for t in os.listdir(d):
                 if len(t)>self.shorty_length:
-                    objects.add(self._get_id_from_dirpath(d))
-                else:
+                    if self._get_id_from_dirpath(d) not in objects:
+                        objects.add(self._get_id_from_dirpath(d))
+                        yield self._get_id_from_dirpath(d)
+                elif os.path.isdir(os.path.join(d, t)):
                     paths.append(os.path.join(d, t))
             if paths:
                 d = paths.pop()
             else:
                 d =False
-        return objects
 
     def _create(self, id):
         """
@@ -378,6 +345,38 @@ class PairtreeStorageClient(object):
             raise ObjectNotFoundException
         return [x for x in os.listdir(dirpath) if len(x)>self.shorty_length]
 
+    def isfile(self, id, filepath):
+        """
+        Returns True or False depending on whether the path is a file or not.
+        
+        If the file doesn't exist, False is returned.
+        
+        @param filepath: Path to be tested
+        @type filepath: Directory path
+        @returns: L{bool}
+        """
+        dirpath = os.path.join(self._id_to_dirpath(id), filepath)
+        try:
+            return os.path.isfile(dirpath)
+        except OSError:
+            return False
+        
+    def isdir(self, id, filepath):
+        """
+        Returns True or False depending on whether the path is a subdirectory or not.
+        
+        If the path doesn't exist, False is returned.
+        
+        @param filepath: Path to be tested
+        @type filepath: Directory path
+        @returns: L{bool}
+        """
+        dirpath = os.path.join(self._id_to_dirpath(id), filepath)
+        try:
+            return os.path.isdir(dirpath)
+        except OSError:
+            return False
+
     def put_stream(self, id, path, stream_name, bytestream, buffer_size = 1024 * 8):
         """
         Store a stream of bytes into a file within a pairtree object.
@@ -396,6 +395,7 @@ class PairtreeStorageClient(object):
         @param buffer_size: (Optional) Used for streaming filelike objects - defines the size of the buffer
         to read in each cycle.
         @type buffer_size: integer
+        @returns: tuple C{(hashing_algorithm, hash)} or None if hashing is disabled
         """
         dirpath = os.path.join(self._id_to_dirpath(id))
         if path:
@@ -403,6 +403,8 @@ class PairtreeStorageClient(object):
         if not os.path.exists(dirpath):
             os.makedirs(dirpath)
         f = open(os.path.join(dirpath, stream_name), "wb")
+        if self.hashing_type != None:
+            hash_gen = getattr(hashlib, self.hashing_type)()
         try:
             # Stream file-like objects in with buffered reads
             if hasattr(bytestream, 'read'):
@@ -411,11 +413,43 @@ class PairtreeStorageClient(object):
                 chunk = bytestream.read(buffer_size)
                 while chunk:
                     f.write(chunk)
+                    if self.hashing_type != None:
+                        hash_gen.update(chunk)
                     chunk = bytestream.read(buffer_size)
             else:
                 f.write(bytestream)
+                if self.hashing_type != None:
+                    hash_gen.update(bytestream)
         finally:
             f.close()
+        
+        if self.hashing_type != None:
+            return (self.hashing_type, hash_gen.hexdigest())
+
+    def get_appendable_stream(self, id, path, stream_name):
+        """
+        Reads a filehandle for a pairtree object. This is a "wb+" opened file and
+        so can be appended to and obeys 'seek'
+        
+        >>> with store.get_appendable_stream('foobar:1','data/images', 'image001.tif') as stream:
+                # Do something with the C{stream} handle
+                pass
+
+        stream is closed at the end of a C{with} block
+
+        @param id: Identifier for the pairtree object to read from
+        @type id: identifier
+        @param path: (Optional) subdirectory path to retrieve file from
+        @type path: Directory path
+        @param stream_name: Name of the file to read in
+        @type stream_name: filename
+        @returns: L{file}
+        """
+        file_path = os.path.join(self._id_to_dirpath(id), stream_name)
+        if path:
+            file_path = os.path.join(self._id_to_dirpath(id), path, stream_name)
+        f = open(file_path, "wb+")
+        return f
 
     def get_stream(self, id, path, stream_name, streamable=False):
         """
@@ -469,7 +503,10 @@ class PairtreeStorageClient(object):
             file_path = os.path.join(self._id_to_dirpath(id), path, stream_name)
         if not os.path.exists(file_path):
             raise PartNotFoundException(id=id, path=path, stream_name=stream_name,file_path=file_path)
-        os.remove(file_path)
+        if os.path.isdir(file_path):
+            os.rmdir(file_path)
+        else:
+            os.remove(file_path)
 
     def del_path(self, id, path, recursive=False):
         """
